((
$Date: 2008-09-27 18:19:05 +0200 (sb, 27 sep 2008) $
$Revision: 586 $
$Author: astrorafael $
$HeadURL: svn://localhost/repos/vfx/forthql/tags/forthql-1.0/src/forthql.fs $

2008 By Rafael Gonzalez Fuentetaja.
This file is placed in the public domain. NO WARRANTY.
))
\ #################################
\ *! forthql
\ *R @node{Forth-SQL Interface}
\ *T Forth-SQL Interface
\ #################################

\ ***************
\ *S Introduction
\ ***************

\ *P This implementation of ForthQL depends on a two basic principles
\ ** of operation:
\ *(
\ *B A simple API that hides the details of DBMS connections.
\ ** Connection/Disconnection procedures are out of the scope of ForthQL.
\ *B A Callback interface can be established
\ ** with the DBMS to process the query result set row by row.
\ *)

\ ==============
\ *N Portability
\ ==============

\ *P This module is written in *\i{VFX Forth for Linux} but it can probably
\ ** be ported easily to other Forth environments as well.

\ *P The following aspects of VFX has been used:
\ *(
\ *B Exception messages and codes.
\ *B Modules, encapsulating words into modules and exporting words.
\ *B GenIO Architecure. Memory Device.
\ *B MPEisms. Convenience words defined by MPE for its products
\ ** like *\fo{?comp}, *\fo{$CRLR}, *\fo{ALLOT&ERASE} or *\fo{?throw}
\ *)


\ *********************
\ *S Embedding SQL code
\ *********************

\ *P Examples shown are based on the paper by N. J. Nelson cited in the
\ ** *\f{References} section.

\ *P The main idea of this interface is to embed multiline SQL statements
\ ** inside colon definitions as shown below:

\ *E : TEST1
\ **   SQL| INSERT INTO delegates      \ example comment 1
\ **   VALUES ('Chuck','Moore',1) |SQL
\ ** ;

\ *P SQL statements are started by the word *\fo{SQL|} and ended either with
\ ** *\fo{|SQL}, *\fo{|SQL.} or *\fo{|SQL>} .
\ ** They span one or more lines and Forth comments starting
\ ** with *\fo{\} can appear in between. These comments are
\ ** typically used to document the SQL code itself.

\ *P The SQL statement is compiled into a Z-string inside the current
\ ** definition (*\fo{TEST1} in this case) and sent to the DBMS for execution
\ ** when the current definition is executed.
\ ** Apart from these comments, which are stripped in the actual SQL string,
\ ** the syntax of the SQL statement depends on the actual DBMS being used.

\ *P Ending word variations *\fo{|SQL} or *\fo{|SQL>} are used whenever zero or
\ ** one Forth callback action are used. See glossary section below.

\ *P Lack of SQL ending word inside a file will result in "delimiter not found"
\ ** exception.

\ *******************
\ *S Dynamic SQL code
\ *******************

\ *P ForthQL allows not only static SQL statements to be embedded in Forth
\ ** but also allows for dynamic SQL statement generation by means of
\ ** SQL parameters as in the example:

\ *E : TEST2 					 
\ **   SQL| SELECT cuscode,cusname FROM customers  \ Fixed part
\ **   WHERE cusid BETWEEN
\ **   | LOWLIMIT |  AND | HIGHLIMIT               \ parameters
\ **   |SQL> TESTOUT                               \ Output
\ ** ;

\ *P In this example, we build a template SQL statement.
\ ** Words *\fo{LOWLIMIT} and *\fo{HIGHLIMIT} are template
\ ** parameters and will patch the SQL string with proper values
\ ** at runtime, when *\fo{TEST2} is executed.

\ *P *\fo{CHAR |} is both used as the ending character of the static
\ ** SQL string (as in the first and third appearance) and as
\ ** a Forth word that restarts the SQL string compilation
\ ** (as in the second appearance).

\ *P SQL parameters are specified like in the example below:

\ *E
\ ** :noname    \ ca1 u1 --
\ **   s"    12" drop -rot move ;
\ ** 5 CHARS SQLParameter: LOWLIMIT
\ **
\ **:noname    \ ca1 u1 --
\ **    s" 100009" drop -rot move ;
\ ** 6 CHARS SQLParameter: HIGHLIMIT

\ *P At compile time, when the template SQL is being built, parameters
\ ** need to specify how many space or 'width' they need in the SQL string.
\ ** *\fo{LOWLIMIT} needs *\fo{5 CHARS} and
\ ** *\fo{HIGHLIMIT} needs *\fo{6 CHARS}.

\ *P At runtime, SQL parameters run an execution token *\i{xt}
\ ** This token is for a word that must fill in just the part of the
\ ** SQL string *\i{ca1 u1} reserved to its parameter. Of course,
\ ** *\i{u1} is equal tothe parameter 'width'. Nameless definitons above
\ ** will fill reserved spaces with literals *\f{   12} and *\f{100009}
\ ** respectvely.

\ *P This mechanism is totally independent from the usual SQL run time
\ ** parameter binding using wildcards.

\ ============================
\ *N Parameter edition support
\ ============================

\ *P The above mechanisms are enough to edit any string needed into the
\ ** parameter memory area. However, we can take advantage of VFX GenIO
\ ** Architecture and use
\ ** well known words like *\fo{TYPE}, *\fo{.} , *\fo{EMIT} and friends.
\ ** The advantages are best appreciated when formatting numbers.
\ ** The above example can be re-written as:

\ *E
\ ** :noname				  \ ca1 u1 --
\ **   [sqlio 12 . sqlio] ;
\ ** 5 CHARS SQLParameter: LOWLIMIT
\ **
\ ** :noname			          \ ca1 u1 --
\ **   [sqlio 100009 . sqlio] ; 
\ ** 8 CHARS SQLParameter: HIGHLIMIT

\ *P Words *\fo{[sqlio} opens a context where standard I/O is redirected
\ ** to the *\i{ca1 u1} memory region using the *\fo{SQL-MemDev} GenIO Device.
\ ** Word *\fo{sqlio]} closes this device and restores standard I/O to previous
\ ** values. Raw access to this memory zone is still possible using
\ ** *\fo{IOCTL-GEN}s, but it should not be necessary in this approach.

\ *P *\b{Warnings:}
\ *(
\ *B Take into account that word *\fo{.} emits a final space.
\ *B Due to implementaion issues, reserve 1 additional *\fo{CHAR} to
\ ** your planned parameter width.
\ *)

\ ************************
\ *S Processing SQL output
\ ************************

\ *P Word *\fo{|SQL} just executes the SQL code for a statement that does not
\ ** produce a *\i{result set}, like an insertion into a table
\ ** (as in *\fo{TEST1}).
\ ** Of course, other situations will result in the retrieval of one or more
\ ** rows in a result set, which can be processed one by one by a Forth word.

\ *P Word *\fo{|SQL>} - as shown in *\fo{TEST2} - will also compile
\ ** some runtime code
\ ** and the execution token for the next word (*\fo{TESTOUT} in this case).
\ ** The runtime code is responsible to invoke *\fo{TESTOUT}
\ ** for each row being retrieved. Some DBMS APIs like SQLite3 has a C callback
\ ** interface designed for this purpose. See the proper section for details.

\ ==================
\ *N Support for OOP
\ ==================

\ *P Extending the concept of *\fo{|SQL>}, another word named *\fo{|SQL>>}
\ ** lets you specify both a callback and a client data pointer.
\ ** This gives some support to integrate callback actions with object methods.

\ *P This feature is highly dependant on the DBMS API and the OOP library used.
\ ** The following example is done using SQLite3 and GForth 'objects.fs'
\ ** library ported to VFX Forth for Linux.

\ *E
\ ** sql-dry-run off
\ **
\ ** : quote   ( -- ) [char] ' emit ;
\ **
\ ** : [qtype]   ( ca u -- )
\ **    postpone quote
\ **    postpone type
\ **    postpone quote
\ ** ; immediate
\ ** 
\ ** : creation
\ **    SQL| CREATE TABLE IF NOT EXISTS Person(
\ **       name TEXT PRIMARY KEY, surname TEXT, age INTEGER);
\ **    |SQL.
\ ** ;
\ ** 
\ ** : persons
\ **    SQL| SELECT * FROM Person; |SQL.
\ ** ;
\ ** 
\ ** object class 
\ **    cell% inst-var m-name
\ **    cell% inst-var m-surname
\ **    cell% inst-var m-age
\ ** end-class Person
\ *E 
\ ** Person methods
\ ** protected
\ **   :noname			          \ ca1 u1 --
\ **     [sqlio m-name $@ [qtype] sqlio] 
\ **   ; 16 chars SQLParameter: /name/
\ ** 
\ **   :noname			          \ ca1 u1 --
\ **     [sqlio m-surname $@ [qtype] sqlio] 
\ **   ; 16 chars SQLParameter: /surname/
\ ** 
\ **   :noname			          \ ca1 u1 --
\ **     [sqlio  m-age ? sqlio] 
\ **   ; 8 chars SQLParameter: /age/
\ ** 
\ **   :m (refresh)  ( colValue** this -- )
\ **     0 sql3-$@ evaluate m-age !
\ **   ;m
\ ** 
\ **   :noname   ( object* nCols colValue** colName** -- ior ) 
\ **     drop nip swap (refresh)
\ **     SQLITE_OK
\ **   ; SQLite3Callback: <<refresh>>
\ *E 
\ ** public
\ **    m:   ( ca1 u1 ca2 u2 age this -- ) \ overrides construct
\ **     0 dup m-name ! m-surname !
\ **     m-age ! m-surname $! m-name $!   
\ **   ;m overrides construct
\ **   
\ **   :m save-person   ( this -- )
\ **      SQL| INSERT OR REPLACE INTO Person
\ **      VALUES( | /name/ |  , | /surname/ | , | /age/ | );
\ **      |SQL.
\ **   ;m
\ **   
\ **   :m refresh ( this -- )
\ **     ." Refresing age from database for " m-name $@ type cr
\ **     SQL| SELECT age FROM Person WHERE name = | /name/
\ **     |SQL>> this <<refresh>> 
\ **   ;m
\ ** 
\ ** end-methods
\ ** 
\ ** : Homer   s" Homer" ;
\ ** : Maggie  s" Maggie" ;
\ ** : Bart    s" Bart" ;
\ ** : Simpson s" Simpson" ;
\ ** 
\ ** Homer  Simpson 43 Person heap-new constant p1
\ ** Maggie Simpson 40 Person heap-new constant p2
\ ** Bart   Simpson 14 Person heap-new constant p3
\ **
\ ** s" simpson.db" db-open db-throw creation
\ **
\ ** p1 save-person
\ ** p2 save-person
\ ** p3 save-person
\ ** p1 refresh

\ *********************
\ *S Debugging SQL code
\ *********************

\ *P *\fo{VARIABLE} *\fo{SQL-DRY-RUN} controls the behaviour of *\fo{|SQL},
\ ** *\fo{|SQL>} and *\fo{|SQL.}
\ ** If set, it will compile code to type the contents of the SQL buffer
\ ** and then *\fo{EXIT}. Parameters are seen with its run-time values
\ ** into their places. Memory overruns in the SQL buffer are a likely cause
\ ** of SQL syntax errors.

\ *P This is a compilation flag. To disable it, you must turn it off
\ ** and reload your code.


\ ******************************
\ *S Tunning the SQL buffer size
\ ******************************

\ *P SQL statements are compiled (as Z strings) inside the current definition
\ ** in a similar way as words *\fo{S"} or *\fo{."} do.
\ ** The size of this *\b{per-word SQL Buffer} is is controlled by the VALUE
\ ** *\fo{#SQLBuffer} with a default size. However, the user can
\ ** change this value prior to defining a given SQL statement to tune for
\ ** very large statements and to avoid waste of memory space for short
\ ** statements.

\ *P A buffer too short for a given statement will result in a
\ ** buffer overflow exception. Forth comments do not count against this limit.

\ *P Words *\fo{SQLBuffer>} and *\fo{>SQLBuffer}, used as a pair surronding
\ ** the definition, change and restore to a previous value
\ ** the SQL buffer size.

\ *P *\fo{VARIABLE} *\fo{SQL-TUNNING} controls the printing of compile-time
\ ** diagnostics on the SQL buffer memory usage. By default it is turned off.

\ *E
\ ** SQL-TUNNING ON
\ ** #148 CHARS SQLBuffer>
\ ** : MYDEF   ( -- )
\ **   SQL| CREATE TABLE IF NOT EXISTS Person(
\ **      oid INTEGER PRIMARY KEY,		\ object Id
\ **      name TEXT,				
\ **      surname TEXT,
\ **      age INTEGER
\ **      );
\ **      |SQL.
\ ** ;
\ ** >SQLBuffer
\ ** \ and the output is ...
\ ** MYDEF SQL Buffer = 148 bytes , used = 148 , wasted = 0

\ *P Example above shows the exact amount of tweaking using *\fo{SQLBuffer>},
\ ** hence the 0 wasted bytes.

\ *****************
\ *S Multithreading
\ *****************

\ *P ForthQL depends on a simple API *\fo{DB-EXECUTE} or *\fo{DB-PROCESS}
\ ** This API hides the database connections.
\ ** It is up to the actual Forth DBMS driver to make this thread-safe
\ ** For instance, it could declare a *\fo{USER} variable *\fo{DB-HANDLE}.
\ ** to have several threads with its own connection.

\ *P Some data structures like *\fo{SQL-MemDev} are conditionally compiled
\ ** into *\fo{USER} variables if *\fo{SQL-MULTITHREAD} is non-zero.

\ ==============
\ *S Limitations
\ ==============

\ *(

\ *B As stated elsewhere, this word set operates on one DB connection or
\ ** at most one DB connection per thread.

\ *B There could be *\b{race conditions} if two threads execute
\ ** *\b{the same dynamic SQL query}
\ ** like shown in *\fo{TEST2}, giving unpredictable results.
\ ** Not only the SQL string buffer must should protected during edition
\ ** itself at runtime, it should be locked until the SQL statement has been
\ ** executed.
\ ** Solving this problem transparently to the user for this unlikely
\ ** situation adds a lot of complexity.
\ ** The simplest workaround is to duplicate the
\ ** SQL statement in two different words (i.e. *\fo{TEST2} and *\fo{TEST2'})
\ ** and let the threads execute each one of them.

\ *B *\fo{SQL-RESULT} for SQLite3 is not as nice as presented in Nelson's
\ ** paper. 
\ ** The Forth callback does not know when is the last time
\ ** to be invoked and this prevents further clean-up actions.

\ *B Defining a maximun field width for a *\fo{SQLParameter:} at compile time
\ ** is not very flexible. Text like fields vary much in their contents and
\ ** you must set an upper limit to them. Finding such limit is a compromise.

\ *)

\ ***************
\ +S Design notes
\ ***************

\ ===========
\ +N DBMS API
\ ===========

\ +P In a previous version, I implemented all the DBMS API words as
\ +* execution vectors. Latter, I realized that this was not necessary.
\ +* Vectored execution is better suited to changes at run-time,
\ +* not compile-time.
\ +* So I just simply introduced dependencies with *\fo{[undefined]}.

\ =============
\ +N SQL Buffer
\ =============

\ +P Using an per-word SQL buffer embedded in the word itself is easier to code
\ +* than introducing a separate pool of memory that needs to be managed.
\ +* Skipping this block of data at runtime is done by surronding it with
\ +* *\fo{AHEAD} and *\fo{THEN}. The compiler then introduces the proper jump.

\ +P The semi-static SQL Buffer idea introduces its own complexity by all this
\ +* buffer adjustment stuff. However, it is the only way I have found to work
\ +* with SQL dynamic parameters. At compile time, a SQL parameter needs
\ +* to compile the right address for the edition at run-time. If I used
\ +* a temporary buffer, then copy to the word, the compiled address would
\ +* be of that temporary buffer.


\ ***********
\ *S Glossary
\ ***********

[undefined] textbuff-sid [if]
  cr .( Include 'Buffer.fth' file for the memory GenIO driver.) cr
     .( Usually located at '/usr/share/doc/VfxForth/Lib/Genio/Buffer.fth') cr
  quit
[then]


[undefined] db-execute
[undefined] db-process      or
[undefined] db-process-with or
[undefined] db-print        or
[undefined] db-throw        or
[if]
   cr .( Need a real DBMS Interface. Load the appropiate Forth driver !)
   quit
[then]



\ ================
\ *N Support words
\ ================

\ *P General purpose words which could be placed elsewhere and not related
\ ** to *\fo{ForthQL}.
\ ** They may be placed in my *\i{Extras} project in a future.

\ --------------------
\ *H Caddr/len strings
\ --------------------

[undefined]  STRING/ [IF]
: STRING/ \ ca1 u1 u -- ca2 u2
\ *G Get the string-matched *\i{ca2 u2} from the string-remaining *\i{ca1 u1}
\ ** and the length *\i{u} of the original string.
\ ** (Jenny Brien).   
   SWAP - TUCK - SWAP
;
[ENDIF]


[undefined] /COMMENT [IF]
: /COMMENT					    \ ca1 u1 -- ca2 u2
\ *G Strip trailing Forth backlash comments from a string.
   tuck   [char] \ scan   rot string/
;
[ENDIF]

\ ----------------------------------
\ *H Input Specification and Parsing
\ ----------------------------------

[undefined]  PARSE-AREA@ [IF]
: PARSE-AREA@             \ -- ca u
\ *G Get the as yet unparsed portion of the input buffer.
\ ** (Jenny Brien).
   SOURCE >IN @ /STRING
;
[ENDIF]


[undefined]  PARSE-AREA! [IF]
: PARSE-AREA!	       \ ca u -- ; 
\ *G Set the portion of the input buffer still to be parsed to *\i{ca u}.
\ ** Must start within the input buffer! (Jenny Brien).  
   DROP SOURCE DROP - 1 CHARS / >IN !
;
[ENDIF]


[UNDEFINED]  PARSE-AREA/ [IF]
: PARSE-AREA/		    \ ca1 u1 -- ca2 u2
\ *G Get the already parsed string *\i{ca2 u2} in the input buffer from the
\ ** yet unparsed *\i{ca1 u1} string. Similar to what *\fo{STRING/} does. 
   nip parse-area@ ROT -
;
[ENDIF]


[UNDEFINED]  SKIP-CHAR  [IF]
: SKIP-CHAR			  \ ca1 u1 --
\ *G Skip *\fo{1 CHARS} (usualy the *\fo{SCAN}ned character)
\ ** from the remaining space *\i{ca1 u1} in the Forth input buffer,
\ ** updating the input buffer.
   1 CHARS /STRING PARSE-AREA!	       
;
[ENDIF]

\ --------------------
\ *H Memory operations
\ --------------------


[UNDEFINED] +MOVE [IF]
: +MOVE					  \ ca1 u1 ca2 u2 -- ca3 u3
\ *G Move memory region *\i{ca2 u2} to receiver memory buffer *\i{ca1 u1}.
\ ** Available receiver memory buffer is now *\i{ca3 u3}.
   dup >r				\ keep moved bytes in Rstack
   2over drop swap move			\ move string to buffer
   r> /string				\ update remaining buffer
;
[ENDIF]


\ =================
\ *N Generic DB API
\ =================

MODULE ForthQL


\ *P This simple API is needed for ForthQL.
\ ** These words must be implemented by the underlying
\ ** DBMS Forth driver module. You must include this driver module before
\ ** *\fo{ForthQL} module. Connection to DBMS is out of the scope.

\ *P Usage of zero-teminated strings (Z-strings) are required to ease
\ ** interfacing with foreign, C-based DBMS APIs.

(( --For DocGen Use only

: db-execute			  \ z-addr -- ior
\ *G Execute SQL statement *\i{z-addr} not returning any output, like
\ ** table creation or row insertion/update/delete.   
\ ** *\i{ior} code signals operation result.         
;

: db-process			  \ z-addr xt -- ior
\ *G Execute SQL statement *\i{z-addr}.
\ ** When complete, word given by *\i{xt} is repetidely called,
\ ** row by row, to process output.      
\ ** *\i{ior} code signals operation result.      
;

: db-process-with			  \ z-addr xt1 xt2 -- ior
\ *G Execute SQL statement *\i{z-addr}.
\ ** When complete, word given by *\i{xt2} is repetidely called,
\ ** row by row, to process output.
\ ** *\i{ior} code signals operation result.    
\ ** *\i{xt1} is the execution token for a word that - when executed -
\ ** return a cell with client data.
\ ** Client data can be anything, but mostly will be an object handle
\ ** for OOP support (the *\fo{this} keyword).   
\ *C <clientData> VALUE clientDataPtr   \ xt1 is ' clientDataPtr.
\ *C : this ( -- handle) ... ;          \ xt1 is ' this   
;

: db-throw		         \ ior --
\ *G Map DB specific *\i{ior} into an appropiate user exception and throw it.
\ ** Exceptions are DBMS specific.
;

: db-print		        \ z-addr --
\ *G Print the SQL statement *\i{z-addr} being executed and its result set
\ ** nicely formatted.   
;
))

\ ========================
\ *H SQL Buffer management
\ ========================

#256 CHARS Value #SQLBuffer
\ *G SQL Memory Buffer default size in bytes.

EXPORT #SQLBuffer

: SQLBuffer>				  \ n1 -- n2
\ *G Set the current SQL buffer size to *\i{n1} bytes.
\ ** Return the old value *\i{n2} for latter restoration.
\ ** Intended to use in pair with *\fo{>SQLBuffer}. See example in
\ ** the *\i{Tunning the SQL buffer size} section.   
   #SQLBuffer swap TO #SQLBuffer
;
EXPORT  SQLBuffer>

: >SQLBuffer				  \ n1 --
\ *G Set the current SQL buffer size to *\i{n1} bytes.  
   TO #SQLBuffer
;

EXPORT  >SQLBuffer

: allot-sql-buffer		        \ -- ca1
\ +G Allot a new SQL buffer of size *\fo{#SQLBuffer} bytes.
\ +* Return the SQL buffer start address.   
   here #SQLBuffer allot&erase 0 c,
;

Variable SQL-TUNNING
\ *G Flag. If true, prints the actual size SQL string being compiled
\ ** so that you can fine tune *\fo{#SQLBuffer} for that word.
SQL-TUNNING OFF

EXPORT SQL-TUNNING

: ?sql-tunning				  \ ca1 u1 -- ca1
\ *G Print a summary report of bytes being used for the SQL Buffer
\ ** in the word being defined.
   sql-tunning @ if
      over ip>nfa .name			\ print current word name
      ." SQL Buffer = " #SQLBuffer .
      ." bytes , used = " dup .
      ." , wasted = " #SQLBuffer swap -  . cr
      exit
   then
   drop
;

EXPORT ?SQL-TUNNING

\ ==========================
\ *N SQL Code Compiler words
\ ==========================

Variable SQL-DRY-RUN
\ *G Activate a mode where only SQL compiler words only compile code to print
\ ** the SQL code. Do not execute anything on the DBMS.
\ ** *\b{Warining:} This is a compile (loadtime) flag, not runtime option

EXPORT SQL-DRY-RUN 
SQL-DRY-RUN OFF


\ *P Exceptions thrown at compile time.

ErrDef SQLDelimErr "No SQL Delimiter | found"
\ *G

ErrDef SQLBuffOvf  "SQL Buffer overflow"
\ *G


: end-sql?				  \ -- ca1 u1 flag
\ +G Look for the SQL end of marker *\fo{CHAR |} in current input stream.
\ +* Flag is *\fo{True} if found.
   parse-area@
   [char] | scan dup 0<>
; 


: +crlf					  \ ca1 u1 -- ca2 u2
\ +G Append a CR+LF to SQL Buffer *\i{ca1 u1}.
\ +* Return remaining buffer*\i{ca2 u2}.
   crlf$ count +move
;


: end-parse-sql				  \ ca1 u1 --
\ +G Manipulate the input stream to set the remaining parse area.
\ +* If SQL delimiter *\fo{|} is alone, skip it, otherwise leave it
\ +* as part of the *\fo{|SQL} or *\fo{|SQL>} words.
\ +*
   over char+ c@			\ get char past the delimiter
   BL <= if
      1 chars /string			\ if marker alone, skip it
   then
   parse-area!
;

: buffer-overflow?				  \  ca1 u1 ca2 u2 -- flag
\ +G Detect buffer overflow if new contents *\i{ca2 u2} 
\ +* does not fit in buffer *\i{ca1 u1}.
\ +* True if overflow.   
   nip < nip 
;

: copy-sql				  \ ca1 u1 ca2 u2 -- ca3 u3
\ +G Append SQL Code string *\i{ca2 u2} to remaining SQL memory buffer
\ +* *\i{ca1 u1}
\ +* stripping backlash Forth comments. 
\ +* Return updated remaining memory buffer *\i{ca3 u3}.
\ +* May throw a *\fo{SQLBuffOvf} exception if not enough space in buffer.
   /comment
   4dup buffer-overflow? SQLBuffOvf ?throw
   +move
;


: parse-sql				  \ ca1 u1 -- ca2 u2
\ +G Parse the input stream and copy contents into SQL buffer *\i{ca1 u1} until
\ +* the final marker character *\fo{|} is found, Multiline processing.
\ +* Throw *\fo{SQLDelimErr} if no final marker is found at end of file.
\ +* Leave remaining SQL buffer *\i{ca2 u2}.   
   BEGIN
      end-sql?				\ Delimiter found ?
   0= WHILE
      PARSE-AREA/			\ get parsed line up to EOL
      copy-sql				\ and copy to SQL Buffer
      +crlf				\ append CR+LF for formatting purposes
      REFILL 0= SQLDelimErr ?throw 	\ get more data or die
   REPEAT
   2dup 2>R				\ backup for input stream adjustment
   PARSE-AREA/				\ get final parsed line up to delimiter
   copy-sql 				\ and copy to SQL buffer
   2R> end-parse-sql			\ final input stream adjustment
; 


: SQL|					  \ -- ca1 u1
\ *G Start SQL string compilation, initializing a SQL memory buffer and
\ ** concatenating verbatim until following *\fo{CHAR |}.
\ ** Any backlashed comments are removed.
\ ** Preserves newline characters.
   ?comp
   postpone AHEAD			\ make mark jump address
   allot-sql-buffer swap     \ swap so as not to mess with addr set by AHEAD
   postpone THEN			\ make resolve jump address
   #SQLBuffer parse-sql
; immediate

EXPORT SQL|

: |					  \ ca1 u1 - ca2 u2
\ *G Resume SQL string compilation to SQL buffer *\i{ca1 u1} until next
\ ** *\fo{CHAR |} is found, as above.
\ ** Leave remaining SQL buffer *\i{ca2 u2}.
   ?comp
   parse-sql
; immediate

EXPORT |


: compile-buffer			  \ ca1 u1 --
\ +G Compile (as Z-string) the used SQL memory buffer into the
\ +* current definition, from the remaining SQL buffer *\i{ca1 u1}.
\ +* This makes the SQL buffer available at runtime
\ +* for further processing by the runtime code compiled by
\ +* *\fo{|SQL}, *\fo{|SQL>} or *\fo{|SQL.}
   #SQLBuffer string/ 			\ -- ca2 u2
   ?sql-tunning				\ consume u2, otherwise wasted.
   postpone literal
;

: ?db-execute				  \ --
\ +G Conditionally compile *\fo{DB-EXECUTE} or code just to type   
\ +* SQL statements to standard output.   
   sql-dry-run @ if
      postpone .z$
   else
      postpone db-execute
      postpone db-throw
   then
;


: |SQL				\ ca1 u1 --
\ *G End current SQL statement compilation.
\ ** Receives as input the remaining SQL buffer *\i{ca1 u1} 
\ ** to find out and compile the SQL buffer starting address.   
\ ** Compile internaly a call to *\fo{DB-EXECUTE} with all necessary
\ ** parameters.
\ ** At run time, *\fo{DB-EXECUTE} is invoked, which will send the
\ ** SQL statement for the DBMS to execute.   
   ?comp
   compile-buffer			\ handle input baffer ca1 u1
   ?db-execute
; immediate

EXPORT |SQL

: ?db-process				  \ --
\ +G Conditionally compile *\fo{DB-PROCESS} or code just to type   
\ +* SQL statements to standard output.
   sql-dry-run @ if
      postpone .z$
      ' drop 		\ discard action
   else
      '  postpone literal		       \ parse "action" and get its xt
   postpone db-process
   postpone db-throw
   then
;

: |SQL>				\ ca1 u1 "action" --
\ *G End current SQL statement compilation and compiles the following "action"
\ ** word.
\ ** Receives as input the remaining SQL buffer *\i{ca1 u1} 
\ ** to find out and compile the SQL buffer starting address.      
\ ** Compile a call to *\fo{DB-PROCESS} with all necessary
\ ** parameters.
\ ** At run time, *\fo{DB-PROCESS} is invoked which  will send the
\ ** SQL statement for the DBMS to execute.    
\ ** For each row returned by the DBMS as the result, it will invoke the xt of
\ ** "action".
   ?comp
   compile-buffer		       \ handle input ca1 u1
   ?db-process
; immediate


EXPORT |SQL>


: ?db-process-with				  \ --
\ +G Conditionally compile *\fo{DB-PROCESS-WITH} or code just to type   
\ +* SQL statements to standard output.
   sql-dry-run @ if
      postpone .z$
      ' drop 				\ discard object
      ' drop				\ discard action 
   else
      '  postpone literal	       \ parse "object" and get its xt
      '  postpone literal	       \ parse "action" and get its xt
      postpone db-process-with
      postpone db-throw
   then
;

: |SQL>>				\ ca1 u1 "object" "action" --
\ *G End current SQL statement compilation and compiles both the following
\ ** "object" and  "action" words.
\ ** Receives as input the remaining SQL buffer *\i{ca1 u1} 
\ ** to find out and compile the SQL buffer starting address.      
\ ** Compile a call to *\fo{DB-PROCESS-WITH} with all necessary
\ ** parameters.
\ ** At run time, *\fo{DB-PROCESS-WITH} is invoked which  will send the
\ ** SQL statement for the DBMS to execute.    
\ ** For each row returned by the DBMS as the result, it will invoke the xt of
\ ** "action" with the xt of "object" as a parameter.   
   ?comp
   compile-buffer		       \ handle input ca1 u1
   ?db-process-with
; immediate

EXPORT |SQL>>


: ?db-print				  \ --
\ +G Conditionally compile *\fo{DB-PRINT} or code just to type
\ +* SQL statements to standard output.
   sql-dry-run @ if
      postpone .z$
   else
      postpone db-print
   then
;

: |SQL.				\ ca1 u1 --
\ *G End current SQL statement compilation.
\ ** Receives as input the remaining SQL buffer *\i{ca1 u1} 
\ ** to find out and compile the SQL buffer starting address.   
\ ** Compile a call to *\fo{DB-PRINT} into the current definition.
\ ** At run-time, *\fo{DB-PRINT} is executed which
\ ** sends the SQL statement to the DBMS for execution and prints
\ ** (a more or less nicely formatted) result set.   
   ?comp
   compile-buffer
   ?db-print
; immediate

EXPORT |SQL.

\ =========================
\ *N Dynamic SQL Statements
\ =========================

: SQLParameter:			 \ xt width -- ; [child] ca1 u1 pfa -- ca2 u2
\ *G Define a *\i{SQL parameter} that will get substituted at runtime
\ ** when the definition containing the SQL code is executed.
\ ** *\i{width} is the 'parameter width', that is the amouts of bytes in the
\ ** SQL string that must be reserved to be patched later on.
\ ** Children words receive the remaining SQL buffer *\i{ca1 u1}
\ ** to just to compile *\i{ca1 width} into the word that is defining the SQL
\ ** statement; along with the *\i{xt} and to calculate the remaining buffer. 
\ ** *\i{ca2 u2}.
\ ** *\i{xt} is the execution token for an action word that performs this
\ ** patching, receiving exactly the given memory area, like:
\ *C : parameter-filler   ( ca1 width -- ) ... ;   
   Create , ,			   \ stores parameter width and action
   immediate				\ children are immediate
 DOES>
  ?comp				      \ children are compile-ony words
  >r					\ -- ca1 u1
  r@ @ over > SQLBuffOvf ?throw		\ check for available size
  over postpone literal			\ compile starting address
  r@ @ postpone literal			\ compile remaining size
  over r@ @ BL fill			\ fill SQL string with blanks
  r@ @ /string				\ adjust remaining SQL string
  r> 1 cells + @ compile,		\ compile xt
;

EXPORT SQLParameter:

\ ===============================
\ *H Support to parameter edition
\ ===============================

\ *P The following words help parameter edition at runtime by doing
\ ** I/O at the memory area to fill. They are not strictly needed but
\ ** they are quite convenient.

(( For DecGen Use only
TextBuff: SQL-MemDev 
\ *G GenIO Memory Device SID.
\ ** The multithreaded version is a *\fo{USER} variable, that must be
\ ** initialized per thread with the phrase *\fo{SQL-MemDev initTextBuffSid}.
))

sql-multithread [IF]
   textbuff-sid +User SQL-MemDev
[ELSE]
   TextBuff: SQL-MemDev 
[ENDIF]

EXPORT SQL-MemDev

: +SQLMemDev				  \ ca1 u1 --
\ *G Open the *\fo{SQL-MemDev} GenIO memory device to edit
\ ** memory zone *\i{ca1 u1}. Standard I/O is redirected to this device.
\ ** Words like *\fo{EMIT}, *\fo{.} or *\fo{TYPE} write on this region.   
   2dup
   SQL-MemDev SetIO				
   True open-gen #-69 ?throw drop  \ open memory device with given area
   BL fill			   \ repair GenIO driver damage (0 fill)
;

EXPORT +SQLMemDev

: patch-blank
\ +G Patch default VFX GenIO Memory Device behaviour to append a zero byte
\ +* after any output operation. Puts a *\fo{BL} instead.   
   0 0 0  ioctl-gen		\ get buffer start address
   0 0 -2 ioctl-gen		\ get write pointer offset
   chars + BL swap c!		\ patch it with a blank
; 

: -SQLMemDev	   \ --
\ *G Close the *\fo{SQL-MemDev} GenIO Memory Device to for parameter edition.
\ ** *\b{Warning:} Standard I/O is not yet restored to previous value.   
   patch-blank
   close-gen  #-62 ?throw
;

EXPORT -SQLMemDev

: [sqlio				  \ ca1 u1 -- R: -- ip-handle op-handle
\ *G Convenience macro for the phrase *\fo{[io +SQLMemDev} .
\ ** Open the *\fo{SQL-MemDev} memory device, configures it to use
\ ** buffer *\i{ca1 u1} and redirects standard I/O to it.
\ ** Intended to use inside a colon definiton at the start of a SQLParameter
\ ** action.
\ ** See the *\fo{Dynamic SQL code} section at the beginning of this chapter.
   postpone [io
   postpone +SQLMemDev
; immediate

EXPORT [sqlio

: sqlio]				  \ R: -- ip-handle op-handle
\ *G Convenience macro for the phrase *\fo{-SQLMemDev io]} .
\ ** Close the *\fo{SQL-MemDev} memory device and restores standard I/O
\ ** to its previous settings.   
\ ** Intended to use inside a colon definiton at the end of a SQLParameter
\ ** action.
\ ** See the *\fo{Dynamic SQL code} section at the beginning of this chapter.
   postpone -SQLMemDev
   postpone io]
; immediate

EXPORT  sqlio]


END-MODULE


\ ======
\ *> ###
\ ======
